# Standard imports
import time
from dataclasses import dataclass

# Third party imports

import jax.numpy as jnp
from jax import Array
import jax

@dataclass
class Sampler:
    k: Array
    x0: Array

    def __post_init__(self):
        if self.k.shape[0] != self.x0.shape[0]:
            raise ValueError('k and x0 must have the same number of rows')

    def initial_state(self, n:int=1, seed=None)->Array:
        return self.state(t=0, n=n, seed = seed)
    

    def state(self, t:int=None, n:int=1, seed=None)->Array:
        seed = seed if seed else int(time.time()*1e7)
        key = jax.random.PRNGKey(seed)
        key, subkey = jax.random.split(key)
        if t is None:
            t = jax.random.randint(subkey, 0, minval=0, maxval=self.k.shape[1])
            key, subkey = jax.random.split(key)
        k0 = jax.random.choice(subkey, self.k[:,t], (n,), replace=True)
        key, subkey = jax.random.split(key)
        x0 = jax.random.choice(subkey, self.x0, (n,), replace=True)
        return jnp.column_stack([k0,x0])


    def path(self, initial_state: Array=None, n:int=1, seed=None)->tuple[Array, Array]:
        '''
        Generates n paths, all starting from the same initial state.
        #TODO: different if n=1
        '''
        seed = seed if seed else int(time.time()*1e7)
        key = jax.random.PRNGKey(seed)
        key, subkey = jax.random.split(key)
        if initial_state is None:
            initial_state = self.initial_state(n=1, seed = subkey[0])[0]
            key, subkey = jax.random.split(key)
        k = self.k[(self.k[:,0]== initial_state[:self.k.shape[2]]).all(axis=1)]
        k = jax.random.choice(subkey, k, (n,), replace=True)
        initial_state = jnp.tile(initial_state, (n,1))
        x0 = initial_state[:,-self.x0.shape[1]:]
        return x0, k

       
  


from dataclasses import dataclass
from typing import Callable

from jax import Array


@dataclass(frozen=True)
class Model:
    '''
    Class that defines a model
    '''
    u: Callable[[Array, Array], Array]
    m: Callable[[Array, Array, Array], Array] 
    T:int


from dataclasses import dataclass
from typing import Callable

import jax
from jax import Array
import jax.numpy as jnp
from jax.tree_util import register_pytree_node_class


@dataclass
@register_pytree_node_class
class Policy:
    params: dict
    nn: Callable[[Array,dict], Array]

    def tree_flatten(self):
        return (self.params, ), self.nn
    
    @classmethod
    def tree_unflatten(cls, nn, params):
        return cls(*params, nn)

    def __call__(self, state:Array)->Array:
        state = jnp.atleast_2d(state)
        return self.nn(state, self.params)
    
    def update(self, params:dict)->None:
        self.params = params
        return self

    @property
    def structure(self):
        return jax.tree_map(lambda x: x.size, self.params)
    
    @property
    def size(self):
        '''
        Returns the number of parameters
        '''
        return jax.tree_util.tree_reduce(lambda x, y: x + y, self.structure)




# Standard imports
from typing import Callable
from dataclasses import dataclass
from typing import Protocol
from functools import partial
import warnings

# Third party imports
import jax
from jax import Array
import jax.numpy as jnp
from jax.tree_util import Partial
import optax
import matplotlib.pyplot as plt
# Local imports
from .model import Model
from .policy import Policy


@partial(jax.jit, static_argnames=['model'])
def evaluate_policy(policy: Callable[[Array], Array],
                    model: Model, 
                    k: Array,
                    x0: Array,
                    ) -> Array:
    '''
    Calculates the value function for each of k=1,...,K paths of 
    exogeneous states, k,  with initial condition of endogeneous
    states, x0.
    
    
    Returns:
    ----------
    Vector of value functions at for each path k:
        \sum_{t=0}^T \beta^t u(action_t, state_t(k))
    '''
    if x0.ndim == 1:
        x0 = x0[jnp.newaxis,:]
    if k.ndim == 2:
        k = k[jnp.newaxis,:,:]
    V, state = jnp.zeros((k.shape[0])), jnp.column_stack([k[:,0], x0])
    return jax.lax.fori_loop(0, model.T+1, 
                             body_fun = Partial(time_iteration, k=k, policy=policy, model=model), 
                             init_val = (V,state))[0]

def time_iteration(t, 
                   x:tuple[Array, Array],
                   policy: Callable[[Array], Array],
                   model:Model,
                   k: Array):
    V, state = x
    action = policy(state)
    V += model.u(state, action)
    state_next = model.m(state, action, k[:,t+1,:])
    return V, state_next

def train_step(policy:Policy,
               model: Model,
               k:Array,
               x0:Array,
               optimizer,
               opt_state
               ):
    '''
    Performs one optimization step to update neural network parameters, params,
    where objective function is -1 times the equal-weighted mean of value
    functions across K different paths of exogeneous states and initial 
    conditions.
    
    Parameters:
    ----------
    params: dictionary of neural network parameters
    u: Instantaneous utility function
    v_T: terminal value function at t=T
    m: Law of motion
    k: K x T+1 x n_exstates array, where n_exstates is the number of 
        exogeneous states
    x0: K x n_edstates array, where n_edstates is the number of 
        endogeneous states
    T: number of periods - 1, where t=0,...,T and t=T is death
    optimizer: optax optimizer object
    opt_state: state of optax optimizer
    
    Returns:
    ----------
    params: updated neutral network parameters
    opt_state: updated state of optimizer
    value: objective function value at previous step
    '''
    value, grads = jax.value_and_grad(lambda params: 
            -jnp.mean(evaluate_policy(policy = Partial(policy.nn, params = params),
                                            model= model,
                                            k = k,
                                            x0 = x0
                                            )
                                )
    )(policy.params)
    updates, opt_state = optimizer.update(grads, opt_state)
    params = optax.apply_updates(policy.params, updates)
    return params, opt_state, value


@dataclass
class TrainResult:
    values:Array

    def plot_convergence(self, ax = None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots()
        ax.plot(self.values, **kwargs)
        ax.set_xlabel('Epoch')
        ax.set_ylabel('Objective Function')
        ax.set_title('Convergence of Objective Function')
        return ax

def train_agnostic(policy: Policy,
                    model: Model,
                    k: Array,
                    x0: Array,
                    batch_size: int = 10,
                    optimizer = optax.adam(1e-4),
                    epochs: int = None,
                    )->tuple[Policy,TrainResult]:
    '''
    Optimizes parameters of a neural network, policy, by performing multiple steps
    of stochastic gradient descent. In each step, a batch of size batch_size is
    used from the path of exogeneous states passed, k. Training completes when
    all batches have been used n_cycles times. 
    
    Parameters:
    ----------
    params: dictionary of initialized neural network parameters
    policy: neural network to optimize
    u: Instantaneous utility function
    v_T: terminal value function at t=T
    m: Law of motion
    k: K x T+1 x n_exstates array, where n_exstates is the number of 
        exogeneous states
    x0: K x n_edstates array, where n_edstates is the number of 
        endogeneous states
    T: number of periods - 1, where t=0,...,T and t=T is death
    batch_size: number of rows of k to pass in each training step
    optimizer: optax optimizer object
    n_cycles: number of times to cycle through exogeneous paths, k
    
    Returns:
    ----------
    params: neural network parameters at final step of optimization
    '''
    
    if k.shape[0] != x0.shape[0]:
        raise ValueError('k and x0 must have the same number of paths')
    if k.shape[1] != model.T + 1:
        raise ValueError('k must have T+1 columns')
    
    epochs = epochs if epochs is not None else k.shape[0]//batch_size
    if epochs*batch_size > k.shape[0]:
        key = jax.random.PRNGKey(0)
        indices = jax.random.randint(key, (epochs*batch_size,), 0, k.shape[0])
        k = k[indices]
        x0 = x0[indices]

    opt_state = optimizer.init(policy.params)
    values = jnp.empty(epochs)


    for n in range(0, epochs*batch_size, batch_size):
        k_batch = k[n:n+batch_size]
        x0_batch = x0[n:n+batch_size]
        policy.params, opt_state, value = train_step(policy = policy, 
                                                model = model,
                                                k = k_batch,
                                                x0 = x0_batch,
                                                optimizer = optimizer, 
                                                opt_state = opt_state)
        values = values.at[n//batch_size].set(value)
        print(f'\rIteration {n//batch_size}, ({n/(epochs*batch_size)*100:.2f} %), value:{-value:.6f}', end = '')
    return policy, TrainResult(-values)





from jax import Array


from .model import Model
from .samplers import Sampler
from .policy import Policy
from .train import evaluate_policy



def evaluate_state(state:Array, policy:Policy, model:Model, sampler:Sampler, n:int=10, seed=None)->Array:
    x0, k = sampler.path(initial_state = state, n=n, seed=seed)
    return evaluate_policy(policy, model, k, x0).mean()